import os

import click
from tokenizers import (Tokenizer, decoders, models, pre_tokenizers,
                        processors, trainers)

BOS = "<s>"
EOS = "</s>"
UNK = "<unk>"
PAD = "<pad>"
MASK = "<mask>"


def text_iterator(docs: list, bs: int):
    for i in range(0, len(docs), bs):
        yield docs[i : i + bs]
    pass


def read_iterator(dpath: str, bs: int):
    with open(dpath, "r") as f:
        text_batch = []
        cnt = 0
        for line in f:
            if cnt == bs:
                yield text_batch
                text_batch = []
                cnt = 0
            text_batch.append(line)
            cnt += 1


@click.command()
@click.option(
    "--data-path", type=str, required=True, help="Path to a training text file."
)
@click.option(
    "--save-dir", type=str, required=True, help="Dir to save trianed tokenizer."
)
@click.option(
    "--batch-size",
    type=int,
    default=32,
    help="Batch size for building tokenizer, smaller value for limited resourced situation but takes more time.",
)
@click.option(
    "--vocab-size",
    type=int,
    default=50000,
    help="Number of unique tokens in the vocabulary.",
)
def train_tokenizer(data_path: str, save_dir: str, batch_size: int, vocab_size: int):
    tokenizer = Tokenizer(models.BPE())
    tokenizer.pre_tokenizer = pre_tokenizers.ByteLevel(add_prefix_space=False)
    tokenizer.decoder = decoders.ByteLevel()

    # Set the training hyperparameters
    trainer = trainers.BpeTrainer(
        vocab_size=vocab_size,
        special_tokens=[BOS, PAD, EOS, UNK, MASK],
        initial_alphabet=pre_tokenizers.ByteLevel.alphabet(),
    )

    # Train it with either files or an iterator:
    tokenizer.train_from_iterator(read_iterator(data_path, batch_size), trainer=trainer)

    # Attach the Roberta Processing
    tokenizer.post_processor = processors.RobertaProcessing(
        sep=(EOS, tokenizer.token_to_id(EOS)), cls=(BOS, tokenizer.token_to_id(BOS))
    )

    # And save it:
    tokenizer.save(os.path.join(save_dir, "tokenizer.json"))



import nltk
from datasets.load import load_dataset
from nltk.tokenize.punkt import PunktSentenceTokenizer
from torch.utils.data import DataLoader
from transformers.models.bart.tokenization_bart_fast import BartTokenizerFast

from engawa.data_collator import DataCollatorForDenoisingTasks


def tokenize_sents(
    sentence_tokenizer: PunktSentenceTokenizer,
    text: str,
    bos_token: str,
    eos_token: str,
) -> str:
    return f"{eos_token}{bos_token}".join(sentence_tokenizer.tokenize(text))


def get_dataloader(
    data_path: str,
    split: str,
    bs: int,
    tokenizer: BartTokenizerFast,
    mask_ratio: float,
    poisson_lambda: float,
    max_length: int,
) -> DataLoader:
    nltk.download("punkt")
    sentence_tokenizer = nltk.data.load("tokenizers/punkt/english.pickle")

    bos_token = tokenizer.bos_token
    eos_token = tokenizer.eos_token
    ds = (
        load_dataset("text", data_files=data_path, split="train")
        .map(
            lambda x: tokenizer(
                tokenize_sents(sentence_tokenizer, x["text"], bos_token, eos_token),
                truncation=True,
                max_length=max_length,
                padding="max_length",
            )
        )
        .remove_columns(["text"])
    )

    assert isinstance(tokenizer.bos_token_id, int)
    dc = DataCollatorForDenoisingTasks(
        tokenizer=tokenizer, mask_ratio=mask_ratio, poisson_lambda=poisson_lambda
    )

    shuffle = True if split == "train" else False
    return DataLoader(ds, batch_size=bs, collate_fn=dc, shuffle=shuffle)



import pytorch_lightning as pl
from transformers.models.bart.modeling_bart import (
    BartConfig, BartForConditionalGeneration)
from transformers.models.bart.tokenization_bart_fast import BartTokenizerFast
from transformers.optimization import (
    AdamW, get_polynomial_decay_schedule_with_warmup)


class EngawaModel(pl.LightningModule):
    def __init__(
        self,
        tokenizer: BartTokenizerFast,
        lr: float,
        weight_decay: float,
        num_warmup_steps: int,
        num_training_steps: int,
        model_type: str,
    ) -> None:
        super().__init__()
        self.tokenizer = tokenizer

        eos_token_id = tokenizer.eos_token_id
        bos_token_id = tokenizer.bos_token_id
        pad_token_id = tokenizer.pad_token_id
        vocab_size = tokenizer.vocab_size
        config = BartConfig.from_pretrained(f"facebook/bart-{model_type}")
        config.update_from_string(
            f"eos_token_id={eos_token_id},bos_token_id={bos_token_id},pad_token_id={pad_token_id},vocab_size={vocab_size}"
        )

        self.bart = BartForConditionalGeneration(config)
        self.lr = lr
        self.weight_decay = weight_decay
        self.num_warmup_steps = num_warmup_steps
        self.num_training_steps = num_training_steps

        self.save_hyperparameters()

    def training_step(self, batch, batch_idx):
        out = self.bart(
            input_ids=batch["input_ids"],
            attention_mask=batch["attention_mask"],
            decoder_input_ids=batch["decoder_input_ids"],
            labels=batch["labels"],
        )
        loss = out.loss
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        out = self.bart(
            input_ids=batch["input_ids"],
            attention_mask=batch["attention_mask"],
            decoder_input_ids=batch["decoder_input_ids"],
            labels=batch["labels"],
        )
        loss = out.loss
        self.log("val_loss", loss)
        return {"loss": loss}

    def configure_optimizers(self):
        no_decay = ["bias", "LayerNorm.weight"]
        bart_params = [
            {
                "params": [
                    p
                    for n, p in self.bart.named_parameters()
                    if not any(nd in n for nd in no_decay)
                ],
                "weight_decay": self.weight_decay,
            },
            {
                "params": [
                    p
                    for n, p in self.bart.named_parameters()
                    if any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.0,
            },
        ]
        optimizer = AdamW(bart_params, lr=self.lr)

        return (
            [optimizer],
            [
                get_polynomial_decay_schedule_with_warmup(
                    optimizer,
                    num_warmup_steps=self.num_warmup_steps,
                    num_training_steps=self.num_training_steps,
                )
            ],
        )



import math
from dataclasses import dataclass
from typing import Dict, List

import numpy as np
import torch
from numpy.random import permutation, poisson
from transformers.tokenization_utils_base import PreTrainedTokenizerBase


@dataclass
class DataCollatorForDenoisingTasks:
    """Data collator used denoising language modeling task in BART.
    The implementation is based on
    https://github.com/pytorch/fairseq/blob/1bba712622b8ae4efb3eb793a8a40da386fe11d0/fairseq/data/denoising_dataset.py.
    The default paramters is based on BART paper https://arxiv.org/abs/1910.13461.
    """

    tokenizer: PreTrainedTokenizerBase
    mask_ratio: float = 0.3
    poisson_lambda: float = 3.5
    permutate_sentence_ratio: float = 1.0
    pad_to_multiple_of: int = 16

    def __post_init__(self):
        if self.tokenizer.mask_token is None or self.tokenizer.eos_token is None:
            raise ValueError

    def __call__(self, examples: List[Dict[str, List[int]]]) -> Dict[str, np.ndarray]:
        """Batching, adding whole word mask and permutate sentences
        Args:
            examples (dict): list of examples each examples contains input_ids field
        """
        # Handle dict or lists with proper padding and conversion to tensor.
        batch = self.tokenizer.pad(
            examples, pad_to_multiple_of=self.pad_to_multiple_of, return_tensors="np"
        )
        batch["decoder_input_ids"] = self.shift_tokens_right(batch["input_ids"])

        do_permutate = False
        if self.permutate_sentence_ratio > 0.0:
            batch["input_ids"] = self.permutate_sentences(batch["input_ids"])
            do_permutate = True

        if self.mask_ratio:
            batch["input_ids"], batch["labels"] = self.add_whole_word_mask(
                batch["input_ids"], do_permutate
            )

        # np array to pt tensor
        batch_pt = {k: torch.from_numpy(v) for k, v in batch.items()}

        return batch_pt

    def shift_tokens_right(self, inputs):
        """Shift decoder input ids right: https://github.com/huggingface/transformers/issues/7961.
        Examples:
            <s>My dog is cute.</s><s>It loves to play in the park.</s><pad><pad>
            shift to -> </s><s>My dog is cute.</s><s>It loves to play in the park.<pad><pad>
        """

        shifted_inputs = np.roll(inputs, 1, axis=-1)

        # replace first token with eos token
        shifted_inputs[:, 0] = self.tokenizer.eos_token_id

        # when there's padding, the last eos tokens will not be rotate to first positon
        # we'll need to replace it with a padding token

        # replace eos tokens at the end of sequences with pad tokens
        end_with_eos = np.where(shifted_inputs[:, -1] == self.tokenizer.eos_token_id)
        shifted_inputs[end_with_eos, -1] = self.tokenizer.pad_token_id

        # find positions where where's the token is eos and its follwing token is a padding token
        last_eos_indices = np.where(
            (shifted_inputs[:, :-1] == self.tokenizer.eos_token_id)
            * (shifted_inputs[:, 1:] == self.tokenizer.pad_token_id)
        )

        # replace eos tokens with pad token
        shifted_inputs[last_eos_indices] = self.tokenizer.pad_token_id
        return shifted_inputs

    def permutate_sentences(self, inputs):
        results = inputs.copy()

        full_stops = inputs == self.tokenizer.eos_token_id

        sentence_ends = np.argwhere(full_stops[:, 1:] * ~full_stops[:, :-1])
        sentence_ends[:, 1] += 2
        num_sentences = np.unique(sentence_ends[:, 0], return_counts=True)[1]
        num_to_permute = np.ceil(
            (num_sentences * 2 * self.permutate_sentence_ratio) / 2.0
        ).astype(int)

        sentence_ends = np.split(
            sentence_ends[:, 1],
            np.unique(sentence_ends[:, 0], return_index=True)[1][1:],
        )

        for i in range(inputs.shape[0]):
            substitutions = np.random.permutation(num_sentences[i])[: num_to_permute[i]]

            ordering = np.arange(0, num_sentences[i])
            ordering[substitutions] = substitutions[
                np.random.permutation(num_to_permute[i])
            ]

            index = 0
            for j in ordering:
                sentence = inputs[
                    i, (sentence_ends[i][j - 1] if j > 0 else 0) : sentence_ends[i][j]
                ]
                results[i, index : index + sentence.shape[0]] = sentence
                index += sentence.shape[0]
        return results

    def add_whole_word_mask(self, inputs, do_permutate):
        labels = inputs.copy()

        special_tokens_mask = [
            self.tokenizer.get_special_tokens_mask(val, already_has_special_tokens=True)
            for val in labels.tolist()
        ]
        special_tokens_mask = np.array(special_tokens_mask, dtype=bool)

        # determine how many tokens we need to mask in total
        is_token = ~(labels == self.tokenizer.pad_token_id) & ~special_tokens_mask
        num_to_mask = int(math.ceil(is_token.astype(float).sum() * self.mask_ratio))
        if num_to_mask == 0:
            return inputs, labels

        # generate a sufficient number of span lengths
        lengths = poisson(lam=self.poisson_lambda, size=(num_to_mask,))
        while np.cumsum(lengths, 0)[-1] < num_to_mask:
            lengths = np.concatenate(
                [lengths, poisson(lam=self.poisson_lambda, size=(num_to_mask,))]
            )

        # remove all spans of length 0
        # Note that BART inserts additional mask tokens where length == 0,
        # which we do not implement for now as it adds additional complexity
        lengths = lengths[lengths > 0]

        # trim to about num_to_mask tokens
        idx = np.argmin(np.abs(np.cumsum(lengths, 0) - num_to_mask)) + 1
        lengths = lengths[: idx + 1]

        # select span start indices
        # print("IS TOKEN")
        # print(is_token)
        # print(sum(list(map(lambda x: 1 if(x) else 0, is_token[0]))))
        token_indices = np.argwhere(is_token == 1)
        # print("TOKEN INDICES")
        # print(token_indices)
        span_starts = permutation(token_indices.shape[0])[: lengths.shape[0]]

        # prepare mask
        masked_indices = np.array(token_indices[span_starts])
        # print("MASKED INDICES")
        # print(masked_indices)
        mask = np.full_like(labels, fill_value=False)

        # mask span start indices
        for mi in masked_indices:
            mask[tuple(mi)] = True
        lengths -= 1

        # fill up spans
        max_index = labels.shape[1] - 1
        remaining = (lengths > 0) & (masked_indices[:, 1] < max_index)
        while np.any(remaining):
            masked_indices[remaining, 1] += 1
            for mi in masked_indices:
                mask[tuple(mi)] = True
            lengths -= 1
            remaining = (lengths > 0) & (masked_indices[:, 1] < max_index)

        # place the mask tokens
        mask[np.where(special_tokens_mask)] = False
        inputs[np.where(mask)] = self.tokenizer.mask_token_id

        if not do_permutate:
            labels[np.where(mask)] = -100
        else:
            labels[np.where(special_tokens_mask)] = -100

        # remove mask tokens that are not starts of spans
        to_remove = (mask == 1) & np.roll((mask == 1), 1, 1)
        new_inputs = np.full_like(labels, fill_value=self.tokenizer.pad_token_id)

        # splits = list(map(lambda x: x.reshape(-1),  np.split(inputs_copy, indices_or_sections=2, axis=0))
        for i, example in enumerate(
            np.split(inputs, indices_or_sections=new_inputs.shape[0], axis=0)
        ):
            new_example = example[0][~to_remove[i]]
            new_inputs[i, 0 : new_example.shape[0]] = new_example

        # batching now fixed
        return new_inputs, labels



import click
import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers.csv_logs import CSVLogger
from pytorch_lightning.loggers.wandb import WandbLogger
from transformers.models.bart.tokenization_bart_fast import BartTokenizerFast

from engawa.data_loader import get_dataloader
from engawa.model import EngawaModel


@click.command()
@click.option(
    "--tokenizer-file", type=str, required=True, help="Path to a tokenizer file."
)
@click.option(
    "--train-file",
    type=str,
    required=True,
    help="Line separated text file for training.",
)
@click.option(
    "--val-file",
    type=str,
    required=True,
    help="Line separated text file for validation.",
)
@click.option(
    "--default-root-dir", type=str, required=True, help="Path to save generated files."
)
@click.option(
    "--ckpt-path",
    type=str,
    required=False,
    default=None,
    help="If given, start from that checkpoint.",
)
@click.option(
    "--wandb-proj-name",
    type=str,
    required=False,
    default=None,
    help="If given, log to that wandb project.",
)
@click.option(
    "--val-check-interval",
    type=float,
    required=False,
    default=0.25,
)
@click.option(
    "--max-steps",
    type=int,
    required=False,
    default=500000,
    help="Set -1 for infinite training.",
)
@click.option(
    "--model-type",
    type=click.Choice(
        ["base", "large"],
        case_sensitive=False,
    ),
    required=False,
    default="large",
    help="BART type, `base` or `large`.",
)
@click.option("--seed", type=int, required=False, default=10)
@click.option("--bs", type=int, required=False, default=32)
@click.option("--max-length", type=int, required=False, default=1024)
@click.option("--lr", type=float, required=False, default=0.0004)
@click.option("--weight-decay", type=float, required=False, default=0.01)
@click.option("--num-warmup-steps", type=int, required=False, default=10000)
@click.option("--mask-ratio", type=float, required=False, default=0.3)
@click.option("--poisson_lambda", type=float, required=False, default=3.5)
def train_model(
    tokenizer_file: str,
    train_file: str,
    val_file: str,
    default_root_dir: str,
    ckpt_path: str,
    wandb_proj_name: str,
    val_check_interval: float,
    max_steps: int,
    model_type: str,
    seed: int,
    bs: int,
    max_length: int,
    lr: float,
    weight_decay: float,
    num_warmup_steps: int,
    mask_ratio: float,
    poisson_lambda: float,
):
    tokenizer = BartTokenizerFast(tokenizer_file=tokenizer_file)

    pl.seed_everything(seed)

    train_dl = get_dataloader(
        train_file,
        "train",
        bs,
        tokenizer,
        mask_ratio,
        poisson_lambda,
        max_length=max_length,
    )
    val_dl = get_dataloader(
        val_file,
        "validation",
        bs,
        tokenizer,
        mask_ratio,
        poisson_lambda,
        max_length=max_length,
    )

    if wandb_proj_name is not None:
        logger = WandbLogger(project=wandb_proj_name, save_dir=default_root_dir)
    else:
        logger = CSVLogger(default_root_dir)

    checkpoint_callback = ModelCheckpoint(monitor="val_loss", save_top_k=1, mode="min")

    num_gpus = torch.cuda.device_count()
    devices = None if num_gpus == 0 else num_gpus
    accelerator = "cpu" if num_gpus == 0 else "gpu"

    trainer = pl.Trainer(
        max_epochs=-1,
        max_steps=max_steps,
        val_check_interval=val_check_interval,
        deterministic=True,
        default_root_dir=default_root_dir,
        callbacks=[checkpoint_callback],
        logger=logger,
        accelerator=accelerator,
        devices=devices,
    )

    model = EngawaModel(
        tokenizer,
        lr,
        weight_decay,
        num_warmup_steps,
        max_steps if max_steps > -1 else 500000,
        model_type=model_type,
    )

    if ckpt_path is not None:
        print(f"Resume training from {ckpt_path}...")
    trainer.fit(model, train_dl, val_dl, ckpt_path=ckpt_path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['engawa', 'engawa.misc']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'datasets>=2.8.0,<3.0.0',
 'nltk>=3.8,<4.0',
 'pytorch-lightning>=1.8.6,<2.0.0',
 'sentencepiece>=0.1.97,<0.2.0',
 'sienna>=0.1.5,<0.2.0',
 'tokenizers>=0.13.2,<0.14.0',
 'transformers>=4.25.1,<5.0.0',
 'wandb>=0.13.7,<0.14.0']

entry_points = \
{'console_scripts': ['engawa = engawa.cli:main']}

setup_kwargs = {
    'name': 'engawa',
    'version': '0.1.4',
    'description': '',
    'long_description': '# engawa\n\n**NOT YET FULLY TESTED**\n\nA simple implementation to pre-train BART from scratch with your own corpus.\n\n\n# Usage\n\nSoon, I will make this pip-installable with CLI commands but at the moment, you need to run it as a repository.\n\n## Installation\n\n```bash\npip install engawa\n```\n\n## Build tokenizer\n\n```bash\nengawa train-tokenizer --data-path /path/to/train.txt --save-dir /path/to/save\n\n# Checkout other options by\nengawa train-tokenizer --help\n```\n\n## Pre-train BART\n\n```bash\nengawa train-model --tokenizer-file /path/to/tokenizer.json --train-file /path/to/train.txt --val-file /path/to/val.txt --default-root-dir /path/to/save/things\n\n# Checkout other options by\nengawa train-model --help\n```\n',
    'author': 'sobamchan',
    'author_email': 'oh.sore.sore.soutarou@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10.8,<4.0.0',
}


setup(**setup_kwargs)



from argparse import ArgumentParser

import numpy as np
import sienna
import torch
from numpy import dot
from numpy.linalg import norm

from engawa.model import EngawaModel


def cossim(a: np.ndarray, b: np.ndarray) -> float:
    return dot(a, b) / (norm(a) * norm(b))


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--ckpt-path", type=str, required=True, help="Path to the pl checkpoint."
    )
    parser.add_argument(
        "--query-text",
        type=str,
        required=True,
        help="Query text to sort other texts by similality.",
    )
    parser.add_argument(
        "--pool-text-file",
        type=str,
        required=True,
        help="Line sep text file for testing.",
    )
    args = parser.parse_args()

    with torch.no_grad():
        target_texts = sienna.load(args.pool_text_file)
        model = EngawaModel.load_from_checkpoint(args.ckpt_path)
        bart = model.bart.model
        bart.eval()
        tok = model.tokenizer

        query_vec = (
            bart(
                **tok(
                    args.query_text, padding=True, truncation=True, return_tensors="pt"
                )
            )
            .last_hidden_state[0, 0, :]
            .numpy()
        )
        target_vecs = (
            bart(
                **tok(target_texts, padding=True, truncation=True, return_tensors="pt")
            )
            .last_hidden_state[:, 0, :]
            .numpy()
        )

        cossims = []
        for i in range(len(target_texts)):
            cossims.append(cossim(query_vec, target_vecs[i, :]))

        print(cossims)



import click

from engawa.tokenizer import train_tokenizer
from engawa.train import train_model


@click.group()
def cli():
    pass


cli.add_command(train_model)
cli.add_command(train_tokenizer)


def main():
    cli()
